/*  ud - universal dump
    Copyright (C) 2013  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    source repo: svn://repo.hu/random
    contact: ud@igor2.repo.hu
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "genusin.h"
#include "help.h"

#define DATA_COLS_AUTO -1024

typedef enum {
	FMT_NONE,
	FMT_INVALID,
	FMT_BIN,
	FMT_OCT,
	FMT_DEC,
	FMT_HEXL,
	FMT_HEXU,
	FMT_HEX = FMT_HEXL
} num_fmt_t;

typedef struct {
	num_fmt_t code;
	char *name;
} num_fmt_name_t;

num_fmt_name_t num_fmt_names[] = {
	{FMT_BIN,  "bin"},
	{FMT_BIN,  "binary"},
	{FMT_BIN,  "2"},
	{FMT_OCT,  "oct"},
	{FMT_OCT,  "octal"},
	{FMT_OCT,  "8"},
	{FMT_DEC,  "dec"},
	{FMT_DEC,  "decimal"},
	{FMT_DEC,  "10"},
	{FMT_HEXL, "hex"},
	{FMT_HEXL, "hexl"},
	{FMT_HEXL, "hexadecimal"},
	{FMT_HEXL, "16"},
	{FMT_HEXU, "HEX"},
	{FMT_HEXU, "hexu"},
	{FMT_HEXU, "HEXADECIMAL"},
	{FMT_NONE, "none"},
	{FMT_NONE, "-"},
	{FMT_NONE, "off"},
	{FMT_INVALID, NULL}
};

num_fmt_t fmt_lookup(const char *fmt_name)
{
	num_fmt_name_t *n;
	for(n = num_fmt_names; n->name != NULL; n++)
		if (strcmp(n->name, fmt_name) == 0)
			return n->code;
	return FMT_INVALID;
}

#define CFG(v) \
	v(int, cols) \
	v(int, *c) \
	v(int, data_cols) \
	v(int, *d) \
	v(int, col_hdr_width) \
	v(int, *H) \
	v(int, word_size) \
	v(int, *w) \
	v(int, printable) \
	v(int, *p) \
	v(int, interactive) \
	v(int, *i) \
	v(int, skip) \
	v(int, *s) \
	v(int, len) \
	v(int, *l) \
	v(dynstr, fmt) \
	v(dynstr, *f) \
	v(dynstr, offs) \
	v(dynstr, *o) \
	v(dynstr, output) \
	v(dynstr, *O) \
	v(dynstr, append) \
	v(dynstr, *A) \

genusin_def_vars(CFG)

num_fmt_t num_fmt;
num_fmt_t offs_fmt;

char offs_fmt_str[16]; /* printf format string for printing an offset (header column) */
char line[1024]; /* line buffer printed again and again preceeded by a \r */
int num_width; /* width of a numeric field */
int verbose, force;
FILE *fdump = NULL;


void print_at(int x, char *s)
{
	int len;
	len = strlen(s);
	if (x+len > cols-1)
		len = cols-1 - x;
	memcpy(line+x, s, len);
}

void flush()
{
	if (interactive) {
		printf("\r%s", line);
		fflush(stdout);
	}
}

void new_line(void)
{
	static int first_line = 1;
	if (!first_line) {
		if (interactive)
			printf("\n");
		else
			printf("%s\n", line);
		if (fdump != NULL)
			fprintf(fdump, "%s\n", line);
	}
	else
		first_line = 0;
	memset(line, ' ', cols-1);
	line[cols] = '\0';
}

void tobin(char *out, unsigned int val)
{
	int n;
	for(n = 0; n < 8; n++) {
		out[n] = !!(val % 128) + '0';
		val <<= 1;
	}
	out[8] = '\0';
}

void main_loop(void)
{
	int c, x, offs;
	char pc[2], buff[32];

	pc[1] = '\0';
	x = data_cols;
	offs = -1;
	while((c = getchar()) >= 0) {
		offs++;
		if (skip > 0) {
			if ((interactive) && ((offs % 32) == 0)) {
				printf("\r(skipping %d more bytes)   ", skip);
				fflush(stdout);
			}
			skip--;
			continue;
		}
		x++;
		if (x > data_cols) {
			new_line();
			x = 0;
			sprintf(buff, offs_fmt_str, offs);
			print_at(0, buff);
		}
		flush();

		/* print plain text version */
		if (printable) {
			pc[0] = c;
			if ((pc[0] < 32) || (pc[0] >= 127))
				pc[0] = '.';
			print_at(cols-1-data_cols+x, pc);
		}

		/* print numeric variant */
		switch(num_fmt) {
			case FMT_BIN:  tobin(buff, c);            break;
			case FMT_OCT:  sprintf(buff, "%03o", c);  break;
			case FMT_DEC:  sprintf(buff, "%03d", c);  break;
			case FMT_HEXL: sprintf(buff, "%02x", c);  break;
			case FMT_HEXU: sprintf(buff, "%02X", c);  break;
			case FMT_INVALID:
			case FMT_NONE: abort();
		}

		print_at(col_hdr_width+2 + x * (num_width+1) + (x/word_size), buff);

		/* exit when length exceeded */
		if (len > 0) {
			len--;
			if (len == 0) {
				new_line();
				if (verbose)
					printf("# <len>");
				return;
			}
		}
	}

	new_line();
	if (verbose)
		printf("# <eof>");
}

void proc_args(int argc, char *argv[]) {
	int n;

/* defaults */
	cols = 80;
	data_cols = DATA_COLS_AUTO;
	col_hdr_width = 8;
	word_size = 8;
	printable = 1;
	verbose = 0;
	force = 0;
	interactive = 1;
	skip = 0;
	len = -1;
	fmt = NULL;
	offs = NULL;
	append = NULL;
	output = NULL;
	num_fmt = FMT_HEX;
	offs_fmt = FMT_DEC;

	c = &cols;
	d = &data_cols;
	H = &col_hdr_width;
	w = &word_size;
	p = &printable;
	f = &fmt;
	o = &offs;
	i = &interactive;
	s = &skip;
	l = &len;
	A = &append;
	O = &output;

	for(n = 1; n < argc; n++) {
		char *err, *cmd, *ocmd;
		ocmd = cmd = argv[n];
		while(*cmd == '-') cmd++;
		err = NULL;
		if ((strcmp(cmd, "h") == 0) || (strcmp(cmd, "help") == 0))
			help(argv[0]);
		else if ((strcmp(cmd, "v") == 0) || (strcmp(cmd, "verbose") == 0))
			verbose++;
		else if ((strcmp(cmd, "f") == 0) || (strcmp(cmd, "force") == 0))
			force = 1;
		else if (strchr(cmd, '=') != NULL) {
			genusin_parse_equ(CFG, cmd, err);
		}
		else {
			n++;
			genusin_load(CFG, cmd, argv[n], err);
		}
		if (err != NULL) {
			fprintf(stderr, "problem with argument '%s': %s\n", ocmd, err);
			exit(1);
		}
	}
}

void dump_cfg()
{
	genusin_print_vars(CFG, stdout, "# CFG ");
}

int main(int argc, char *argv[])
{
	char *cmd, c;
	int dc;

	cmd = getenv("COLUMNS");
	if (cmd != NULL)
		cols = atoi(cmd);

	proc_args(argc, argv);


	if (cols < 40)
		cols = 40;

	if (cols > sizeof(line)-1)
		cols = sizeof(line)-1;

	/* determine formats */
	if (offs != NULL) {
		offs_fmt = fmt_lookup(offs);
		if (offs_fmt == FMT_INVALID) {
			fprintf(stderr, "Invalid offset format %s\n", offs);
			exit(1);
		}
	}
	if (fmt != NULL) {
		num_fmt = fmt_lookup(fmt);
		if (num_fmt == FMT_INVALID) {
			fprintf(stderr, "Invalid data format %s\n", fmt);
			exit(1);
		}
	}

	/* calculate and configure offset column */
	if (offs_fmt != FMT_NONE) {
		switch(offs_fmt) {
			case FMT_BIN:  fprintf(stderr, "printing offsets in binary is not supported yet\n"); return 1;
			case FMT_OCT:  c = 'o'; break;
			case FMT_DEC:  c = 'd'; break;
			case FMT_HEXL: c = 'x'; break;
			case FMT_HEXU: c = 'X'; break;
			case FMT_INVALID:
			case FMT_NONE: abort();
		}
		sprintf(offs_fmt_str, "%%0%d%c", col_hdr_width, c);
	}
	else
		col_hdr_width = -2;

	/* fix word size */
	if (word_size <= 0)
		word_size = sizeof(line);

	/* init numeric format */
	switch(num_fmt) {
		case FMT_BIN:  num_width = 8; break;
		case FMT_OCT:  num_width = 3; break;
		case FMT_DEC:  num_width = 3; break;
		case FMT_HEXL: num_width = 2; break;
		case FMT_HEXU: num_width = 2; break;
		case FMT_INVALID:
		case FMT_NONE: abort();
	}
	dc = word_size * (cols - 2 - col_hdr_width) / (word_size * num_width + (word_size*(1+printable)) + 1);
	if (verbose > 1) {
		printf("# dc %d (maximum data colums that'd fix the terminal)\n", dc);
		printf("# data_cols %d (data colums selected)\n", data_cols);
	}
	if (dc < 1) {
		fprintf(stderr, "Terminal is not wide enough\n");
		return 1;
	}
	if (data_cols == DATA_COLS_AUTO)
		data_cols = dc;
	else if ((dc < data_cols) && (!force)) {
		fprintf(stderr, "WARNING: can't display %d data colums, displaying only %d\n", data_cols, dc);
		data_cols = dc;
	}

	/* set up dumping */
	if ((append != NULL) && (output != NULL)) {
		fprintf(stderr, "Error: can't append and overwite in the same time; please chose either append or output.");
		return 1;
	}

	if (append != NULL) {
		fdump = fopen(append, "a");
		if (fdump == NULL) {
			perror("Error: can't open dump file for appending");
			return 1;
		}
	}
	if (output != NULL) {
		fdump = fopen(output, "w");
		if (fdump == NULL) {
			perror("Error: can't open dump file for writing");
			return 1;
		}
	}

	if (verbose > 1)
		dump_cfg();

	main_loop();

	if (fdump != NULL)
		fclose(fdump);

	printf("\n");
	return 0;
}
